# -*- coding: UTF-8 -*-
###########################################################################
# Eole NG - 2007  
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr 
#  
# dataproxy.py
#  
# classes permettant de vérifier l'authentification sur différents supports
# (actuellement seulement annuaire LDAP)
#       
###########################################################################

# utile ?
class IAuthProxy:
    """Interface pour un système d'authentification et de récupération des infos utilisateurs.
    Mettre les éventuels commandes de connextion, ... dans __init__.py"""
       
    def get_data(self, identifier, identifier_field):
        """Récupère les données correspondant à identifier
        
        @param identifier: à priori l'identifiant d'un utilisateur dans le système de données
        """
        pass

    def authenticate(self, login, password):
        """Vérifie la validité d'un couple login/password

        @param login: identifiant de l'utilisateur dans ce système
        @param password: mot de passe

        retourne True ou False 
        """
        pass
        
import ldap

class LDAPProxy:
    """récupère les infos d'un utilisateur dans un annuaire LDAP"""
    __implements__ = (IAuthProxy,)

    ignored_attrs = ['userPassword']
    service_name = 'ldap'
    group_attrs = ['sambaGroupType', 'displayName', 'cn', 'objectClass', 'gidNumber', 'mail', 'description', 'niveau']

    def __init__(self, hosts, ports, bases, static_data={}, match_attribute='uid'):
        """
        hosts : liste de serveurs ldap (ou un seul)
        ports : liste des ports correspondants
        """
        self.ldap_servers = []
        if type(hosts) is not list:
            hosts = [hosts]
        if type(ports) is not list:
            ports = [ports]
        if type(bases) is not list:
            bases = [bases]
        for host, port, base in zip(hosts, ports, bases):
            self.ldap_servers.append((host, port, base))
        # on attend des listes de valeur pour hosts  et ports
        # si ancienne configuration, on met les valeurs uniques dans une liste
        self.match_attribute = match_attribute
        self.static_data = static_data
        self.proxy = ldap.open(hosts[0], int(ports[0]))
            
    def _proxy_reload(self, host, port):
        """
        recharge la connexion au ldap
        """
        self.proxy.unbind()
        self.proxy = ldap.open(host, int(port))

    def simple_bind(self, search_result, passwd):
        """
        essaye un bind sur le proxy de travail
        """
        try:
            user_dn = search_result[0][0]
            self.proxy.simple_bind_s(user_dn, passwd)
        except Exception, e:
            success = False
        else:
            success = True
        return success
            
    def authenticate(self, user_id, passwd):
        # gestion multi ldap
        for host, port, base in self.ldap_servers:
            self._proxy_reload(host, port)
            ret = {}
            result = self.proxy.search_s(base, ldap.SCOPE_SUBTREE,'(%s=%s)' % (self.match_attribute, user_id))
            success = self.simple_bind(result, passwd)
            # si l'utilisateur n'est pas authentifié, on essaie le serveur ldap suivant
            if success:
                break

        if success:
            try:
                ## si disponible, renvoi des groupes de l'utilisateur en minuscule
                user_groups = []
                info_groups = {}
                classes = []
                group_filters = "(&(objectclass=sambaGroupMapping)(objectClass=posixGroup)(memberUid=%s))" % user_id
                group_result = self.proxy.search_s(base, ldap.SCOPE_SUBTREE, group_filters, self.group_attrs)
                for gr in group_result:
                    group_cn = gr[1]['cn'][0].lower()
                    user_groups.append(group_cn)
                    # si le groupe est une équipe pédagogique, on ajoute les informations sur la classe en question
                    if gr[1].get('description',[''])[0].startswith('Equipe ') and gr[1]['cn'][0].startswith('profs-'):
                        classes.append(gr[1]['cn'][0].replace('profs-', ''))
                    info_groups[group_cn] = gr[1]
                if classes:
                    # on recherche les classes
                    group_filters = "(&(objectclass=sambaGroupMapping)(objectClass=posixGroup)(|(%s)))" % \
                            ")(".join(["cn=%s" % classe for classe in classes])
                    group_result = self.proxy.search_s(base, ldap.SCOPE_SUBTREE, group_filters, self.group_attrs)
                    for gr in group_result:
                         info_groups[gr[1]['cn'][0].lower()] = gr[1]
            except:
                user_groups = []
                info_groups = {}
        # récupération des données 
        try:
            ret = result[0][1]
            # on supprime certains champs (mot de passe)
            for user_attr in self.ignored_attrs:
                if user_attr in ret:
                    del(ret[user_attr])
            ret['user_groups'] = user_groups
            ret['info_groups'] = info_groups
            ret.update(self.static_data)
        except:
            ret = {}
        # self.proxy.unbind_s()
        return success, ret

    def get_data(self, attributes):
        attr_filter = '(&'
        for attr, val in attributes.items():
            attr_filter += '(%s=%s)' % (attr, val)
        attr_filter += ")"
        # gestion multi ldap
        for host, port, base in self.ldap_servers:
            self._proxy_reload(host, port)
            ret = {}
            result = self.proxy.search_s(base, ldap.SCOPE_SUBTREE, attr_filter)
            if result != {}:
                break
        try:
            ## renvoi des groupes de l'utilisateur en minuscule
            user_groups = []
            info_groups = {}
            assert 'memberUid' in result[0][1]
            group_filters = "(&(objectclass=sambaGroupMapping)(objectclass=posixGroup)(memberUid=%s))" % result[0][1]['memberUid'][0]
            group_result = self.proxy.search_s(self.base, ldap.SCOPE_SUBTREE, group_filters, self.group_attrs)
            for gr in group_result:
                user_groups.append(gr[1]['cn'][0].lower())
                info_groups[gr[1]['cn'][0].lower()] = gr[1]
        except:
            user_groups = []
            info_groups = {}
        # récupération des données 
        # FIXME : utiliser le compte utilisateur ou un compte anonyme ?
        try:
            ret = result[0][1]
            # on supprime certains champs (mot de passe)
            for user_attr in self.ignored_attrs:
                if user_attr in ret:
                    del(ret[user_attr])
            ret['user_groups'] = user_groups
            ret['info_groups'] = info_groups
            ret.update(self.static_data)
            return ret
        except:
            import traceback
            traceback.print_exc()
            return {}
