#! /usr/bin/env python
# -*- coding: utf-8 -*-

###########################################################################
#
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# authserver.py
#
# serveur xmlrpc de gestion de session SSO Eole
#
###########################################################################

import socket, os, shutil, time
import xmlrpclib
from md5 import md5
from glob import glob
from ConfigParser import ConfigParser
# imports twisted
from twisted.python import log
from twisted.internet import reactor
# imports sso eole
import config
from saml_utils import get_metadata
from saml_crypto import recreate_cert
from eolesso.ticket import AppTicket
from eolesso.util import *
from eolesso import ticketcache
from eolesso.ticketcache import InvalidSession

class SSOSessionManager(object):

    def __init__(self,data_proxy, user_infos, parent=None):
        self.address = config.AUTH_SERVER_ADDR
        self.init_caches()
        # proxy d'authentification
        self._data_proxy = data_proxy
        self._parent = None
        self.parent_url = parent
        if parent:
            self._parent = xmlrpclib.ServerProxy('%s/xmlrpc' % parent)
            log.msg("- %s : %s" % (_("Parent server defined"), parent))
        # attributs calculés
        self.user_infos = user_infos
        self.filters = {}
        self.apps = {'http':{'all':{}},
                     'https':{'all':{}},
                     'all':{'all':{}}
                    }
        self.sp_meta = {}
        self.req_attributes = []
        self.opt_attributes = []
        self.attribute_sets = {}
        self.load_filters()
        # chargement des metadata SAML et jeux d'attributs
        self.load_attribute_sets()
        self.init_metadata()

    def init_caches(self):
        # cache des "login tickets" (pour empêcher les navigateurs de 'rejouer' l'authentification avec un cache)
        self.login_sessions = {}
        # cache des "granting tickets (ticket de session sso)"
        self.user_sessions = {}
        # cache des attributs utilisateur
        self.data_cache = {}
        # cache des "application" et "proxy" tickets
        self.app_sessions = {}
        # dictionnaire d'association session utilisateur / tickets
        self.user_app_tickets = {}
        # cache pour la gestion des sessions de proxys (expirent en même temps que la session utilisateur)
        self.proxy_granting_sessions = {}
        # cache pour la gestion des sessions de fédération (fournisseur de service)
        self.sp_sessions = {}
        # cache pour la gestion du single logout (logouts en cours)
        # when single logout requested : we store sso_session and app_session of requester if present
        self.pending_logout = {}

    def check_ticket_issuer(self, ticket_id):
        """vérifie qu'un ticket a été généré par ce serveur
        """
        if ticket_id != '':
            ticket_address = "-".join(ticket_id.split('-')[1:-1])
            if ticket_address == self.address:
                return True
        return False

    def end_session(self, session_id):
        """termine la session d'un utilisateur
        """
        # supression des différents tickets associés
        if session_id in self.user_sessions:
            # recherche des tickets associées à la session
            for ticket in self.user_app_tickets[session_id]:
                # recherche des éventuels PGT associés à cet AppTicket
                if ticket.pgt != None:
                    del self.proxy_granting_sessions[ticket.pgt]
                # recherche des éventuels sessions distantes (saml : idp distant) associées
                if hasattr(ticket,'idp_session_index'):
                    if ticket.idp_session_index in self.sp_sessions:
                        del self.sp_sessions[ticket.idp_session_index]
                # self.user_app_tickets[session_id].remove(ticket)
                del self.app_sessions[ticket.ticket]
            if session_id in self.user_app_tickets:
                del self.user_app_tickets[session_id]
            if session_id in self.data_cache:
                del self.data_cache[session_id]
            if session_id in self.pending_logout:
                del self.pending_logout[session_id]
            del self.user_sessions[session_id]
            log.msg(_("Terminating session %s") % session_id)

    def load_applications(self, cfg):
        """Chargement des applications associées aux filtres
        """
        loaded_apps = []
        for app in cfg.sections():
            try:
                app_infos = dict(cfg.items(app))
                try:
                    port = app_infos['port'] = int(app_infos['port'])
                except:
                    port = app_infos['port'] = 'all'
                scheme = app_infos['scheme']
                if scheme not in ['all', 'both']:
                    if scheme not in self.apps:
                        self.apps[scheme]={'all':{}}
                    app_stores = [self.apps[scheme]]
                elif scheme == 'both':
                    app_stores = [self.apps['http'],self.apps['https']]
                else:
                    # pas de protocole spécifié, on stocke dans tous les protocoles
                    app_stores = self.apps.values()
                # on trie les applications par port et chemin
                for store in app_stores:
                    port_store = store.setdefault(port, {})
                    if app_infos['baseurl'] in port_store:
                        log.msg(_("Warning : baseurl %s already defined for protocol %s (overwritten)") % (app_infos['baseurl'],scheme))
                    port_store[app_infos['baseurl']] = [scheme,
                                                        app_infos['addr'],
                                                        app_infos['typeaddr'],
                                                        app_infos['filter']]
                    loaded_apps.append(app)
            except Exception, e:
                log.msg("! %s %s : %s !" % (_("Error loading application"), app, e))
        return loaded_apps

    def load_filters(self):
        """chargement des filtres de données et applications
        """
        loaded_apps = []
        loaded_filters = []
        for conf_file in glob('%s/*.ini' % config.FILTER_DIR):
            conf_filename = os.path.splitext(os.path.basename(conf_file))[0]
            cfg = ConfigParser()
            cfg.read(conf_file)
            if conf_filename.endswith('_apps') or conf_filename == 'applications':
                # liste des applications associées aux filtres
                loaded_apps.extend(self.load_applications(cfg))
            else:
                # ajout d'un filtre de données
                app_filter = dict( (section, dict(cfg.items(section)))
                                   for section in cfg.sections() )
                self.filters[conf_filename] = app_filter
                loaded_filters.append(conf_filename)
        log.msg("- %s : %s" % (_("Filters loaded"), ", ".join(loaded_filters)))
        log.msg("- %s : %s" % (_("Applications defined"), ", ".join(loaded_apps)))

    def load_attribute_sets(self):
        """chargement des jeux d'attributs pour la fédération d'identité
        """
        attr_sets = {}
        for conf_file in glob('%s/*.ini' % config.ATTR_SET_DIR):
            conf_filename = os.path.splitext(os.path.basename(conf_file))[0]
            cfg = ConfigParser()
            cfg.read(conf_file)
            # ajout des jeux d'attributs par entité
            attr_sets = dict( (section, dict(cfg.items(section)))
                               for section in cfg.sections() )
            self.attribute_sets.update(attr_sets)

    def init_metadata(self):
        # réinitialisation des certificats provenant des metadata
        cert_dir = os.path.join(config.METADATA_DIR,'certs')
        if os.path.isdir(cert_dir):
            shutil.rmtree(cert_dir)
        os.makedirs(cert_dir)
        for conf_file in glob('%s/*.xml' % config.METADATA_DIR):
            sp_ident = os.path.splitext(os.path.basename(conf_file))[0]
            self.load_metadata(sp_ident)
        # recherche des attributs requis/optionnels pour le fournisseur de services
        if os.path.exists(os.path.join(config.METADATA_DIR,'attributes.ini')):
            cfg = ConfigParser()
            cfg.read(os.path.join(config.METADATA_DIR,'attributes.ini'))
            if cfg.has_section('required'):
                self.req_attributes = cfg.items('required')
            if cfg.has_section('optional'):
                self.opt_attributes = cfg.items('optional')
        if self.req_attributes == []:
            self.req_attributes = [('ctemail','Adresse mail académique')]

    def load_metadata(self, sp_ident):
        try:
            sp_meta = get_metadata(sp_ident)
            if 'entityID' in sp_meta:
                sp_ident = sp_meta['entityID']
            if type(sp_ident) == unicode:
                sp_ident = sp_ident.encode(config.encoding)
            f_cert = os.path.join(config.METADATA_DIR,'certs','%s.crt' % sp_ident.replace(os.sep,'_'))
            if not os.path.isfile(f_cert):
                # stockage des certificats sous forme de fichiers pem
                cert_data = None
                if 'SignCert' in sp_meta:
                    cert_data = sp_meta['SignCert']
                elif 'SignCert' in sp_meta.get('SPSSODescriptor',{}):
                    cert_data = sp_meta['SPSSODescriptor']['SignCert']
                elif 'SignCert' in sp_meta.get('IDPSSODescriptor',{}):
                    cert_data = sp_meta['IDPSSODescriptor']['SignCert']
                if cert_data:
                    cert_buffer = recreate_cert(cert_data)
                    # stockage du certificat de signature
                    f=open(f_cert, 'w')
                    f.write(cert_buffer)
                    f.close()
        except Exception, e:
            log.msg("- %s (%s)" % (_("Error fetching SAML metadata for %s : ") % sp_ident, str(e)))
            return {}
        if sp_ident not in self.sp_meta:
            if sp_ident in self.attribute_sets:
                attr_msg = ' (%s)' % _('Custom attribute set')
            else:
                attr_msg = ''
            log.msg("- %s : %s%s" % (_("Service Provider initialized"), sp_ident, attr_msg))
        # mise à jour des métadonnées
        self.sp_meta[sp_ident] = sp_meta
        return sp_meta

    def get_metadata(self, sp_ident):
        if sp_ident in self.sp_meta:
            return self.sp_meta[sp_ident]
        else:
            try:
                sp_meta = self.load_metadata(sp_ident)
            except:
                sp_meta = {}
        return sp_meta

    def init_user_session(self, user_data, username, auth_instant , search_attrs = None):
        sessionID = gen_ticket_id('TGC', self.address)
        if sessionID != "":
            # XXX FIXME : utilisation de la clé de fédération comme nom d'utilisateur CAS ?
            if search_attrs is not None:
                self.user_sessions[sessionID] = (search_attrs, auth_instant)
            else:
                self.user_sessions[sessionID] = (username, auth_instant)
            # stockage des données utilisateur
            self.data_cache[sessionID] = user_data
            # initialisation de la liste des tickets associés
            self.user_app_tickets[sessionID] = []
            reactor.callLater(config.SESSION_TIMER, self.end_session, sessionID)
            log.msg(_("Session initialized (%s)") % username)
        return sessionID

    def authenticate_federated_user(self, attrs, idp_ident, auth_instant, current_session=None):
        # recherche du jeu d'attributs à utiliser
        user_data = {}
        sessionID = ''
        if idp_ident in self.attribute_sets:
            attr_set = self.attribute_sets[idp_ident]
        else:
            attr_set = self.attribute_sets['default']
        search_attrs = {}
        # on parcourt les attributs reçus
        for idp_attr, value in attrs.items():
            if idp_attr in attr_set.keys():
                # correspond à la clé de fédération
                search_attrs[attr_set[idp_attr]] = value[0]
        # tous les attributs doivent correspondre à ceux demandés dans le set d'attributs
        if search_attrs and (len(search_attrs) == len(attr_set)):
            first_fed_key = search_attrs.keys()[0]
            user_data = self._data_proxy.get_data(search_attrs)
            username = user_data[first_fed_key][0]
            if current_session:
                if self.user_sessions[current_session][0] == search_attrs:
                    # session déjà existante pour cet utilisateur
                    return current_session, user_data
            if first_fed_key in user_data:
                # utilisateur local trouvé: création de la session
                sessionID = self.init_user_session(user_data, username, auth_instant, search_attrs)
        return sessionID, user_data

    def authenticate(self, username, password):
        user_data = {}
        sessionID = ''
        if type(username) == unicode: username = username.encode(config.encoding)
        if type(password) == unicode: password = password.encode(config.encoding)

        # on demande en premier au serveur père qu'il vérifie les informations
        if self._parent is not None:
            try:
                valeur = self._parent.authenticate(username, password)
                if len(valeur) == 2: #FIXME : compatibilité avec la version précédente
                    sessionID, user_data = valeur
                    auth_instant = time.time()
                else:
                    sessionID, user_data, auth_instant = valeur
            except Exception, e:
                log.msg(_('Error calling function %s on parent server : %s') % ('authenticate',str(e)))
                pass
        if sessionID != '':
            log.msg(_("Session delivered by parent serveur for %s") % username)
            self.data_cache[sessionID] = user_data
            # session du parent, on crée une session locale 'déportée'
            # on ne gère pas d'expiration sur cette session, la validité sera vérifiée sur le parent
            self.user_sessions[sessionID] = (None,)
            self.user_app_tickets[sessionID] = []
        elif password != '':
            # non authentifié par un parent, on essaye de le faire localement
            try:
                authenticated, user_data = self._data_proxy.authenticate(username, password)
                # on stocke le timestamp de la date d'authentification
                auth_instant = time.time()
            except:
                log.msg(_("! Error accessing Authentication service (%s) !") % self._data_proxy.service_name)
                return '', {}
            if authenticated:
                sessionID = self.init_user_session(user_data, username, auth_instant)
                return sessionID, user_data
            log.msg(_("! Authentication failure : %s !") % username)
            return '', {}
        return sessionID, user_data

    def get_login_ticket(self):
        lt = gen_ticket_id('LT', self.address)
        self.login_sessions[lt] = 'login'

    def validate_session(self, session_id):
        """vérifie un session_id existant"""
        if self.check_ticket_issuer(session_id) and session_id in self.user_sessions:
            return True
        if self._parent is not None:
            try:
                valid_session = self._parent.validate_session(session_id)
            except Exception, e:
                log.msg(_('Error calling function %s on parent server : %s') % ('validate_session',str(e)))
                # XXX FIXME : différencier socket.error ?
                valid_session = False
            if valid_session:
                log.msg(_("User session validated on parent server"))
                # si la session n'est pas présente localement, mais validée par le parent, on l'ajoute (ex : redémarrage du serveur local)
                if not session_id in self.user_sessions:
                    self.user_sessions[session_id] = (None,)
                    self.user_app_tickets[session_id] = []
            return valid_session
        log.msg("! %s : %s !" % (_("Invalid session"), session_id))
        return False

    def verify_session_id(self, session_id):
        """vérifie un session_id"""
        if self.check_ticket_issuer(session_id) and session_id in self.user_sessions:
                return True, session_id
        if self._parent is not None:
            try:
                return self._parent.verify_session_id(session_id)
            except Exception, e:
                log.msg(_('Error calling function %s on parent server : %s') % ('verify_session',str(e)))
                return False, session_id
        log.msg("! %s : %s !" % (_("Unknown user session"),session_id))
        return False, session_id

    def get_user_info(self, session_id, details=False):
        """méthode de convenance pour récupérer l'utilisateur"""
        if session_id in self.user_sessions:
            if self.check_ticket_issuer(session_id):
                if details:
                    return self._get_user_details(session_id, "", {})
                else:
                    return self.user_sessions[session_id][0]
            else:
                try:
                    # session du serveur parent
                    return self._parent.get_user_info(session_id, details)
                except:
                    pass
        raise InvalidSession(session_id)

    def get_auth_instant(self, app_ticket):
        """renvoie la date à laquelle l'authentification a eu lieu
        """
        if app_ticket in self.app_sessions:
            ticket = self.app_sessions[app_ticket]
            session_id = ticket.session_id
        if session_id in self.user_sessions:
            return self.user_sessions[session_id][1]
        raise InvalidSession(session_id)

    def logout(self, session_id):
        """vérifie un session_id"""
        if session_id in self.user_sessions:
            self.end_session(session_id)
            if not self.check_ticket_issuer(session_id):
                # session délivrée par le parent ?
                try:
                    return self._parent.logout(session_id)
                except Exception, e:
                    log.msg(_('Error calling function %s on parent server : %s') % ('logout',str(e)))
                    pass
            else:
                return 'ok'
        return ''

    def get_app_ticket(self, session_id, appurl, ticket_prefix='ST', from_credentials=False):
        # si cette session utilisateur est valide, on crée un ticket pour l'application
        appSessionID = ''
        if self.validate_session(session_id):
            # génération d'un nouveau ticket d'application associé à la session
            appSessionID = gen_ticket_id(ticket_prefix, self.address)
            ticket = self.init_app_session(appSessionID, session_id, appurl, from_credentials)
            if ticket.filter != '':
                filter_msg = ' (%s : %s)' % (_("attribute filter"), ticket.filter)
            else:
                filter_msg = ''
            log.msg("%s %s%s" % (_("Session authorized for service"), ticket.service_url, filter_msg))
            return appSessionID
        else:
            log.msg(_("Unknown user session"))
        log.msg("! %s %s !" % (_("Failed to create session for service"), appurl))
        return ''

    def init_app_session(self, appSessionID, session_id, appurl, from_credentials):
        """génère un ticket d'application et initialise son timeout
        """
        ticket = AppTicket(appSessionID, session_id, appurl, self.address, from_credentials=from_credentials)
        # détection d'un éventuel filtre sur les données utilisateur
        ticket.filter = self._check_filter(ticket.service_url)
        log.msg("%s : %s" % (_("Adding session for service"), ticket.service_url))
        # stockage du ticket
        self.app_sessions[appSessionID] = ticket
        self.user_app_tickets[session_id].append(ticket)
        # mise en place des timeouts de session
        reactor.callLater(config.APP_TIMER, self.invalidate_app_ticket, appSessionID)
        return ticket

    def invalidate_app_ticket(self, appSessionID):
        """invalide un ticket d'application pour empêcher son utilisation
        """
        if appSessionID in self.app_sessions:
            self.app_sessions[appSessionID].valid = False

    def verify_app_ticket(self, app_ticket, appurl):
        """vérifie le ticket fourni par une application"""
        verified = False
        ticket = app_ticket
        if app_ticket in self.app_sessions:
            ticket = self.app_sessions[app_ticket]
            verified = ticket.verif_ticket(app_ticket, appurl)
        else:
            # ticket inconnu, essai de vérification sur un serveur parent
            if self._parent is not None:
                try:
                    return self._parent.verify_app_ticket(app_ticket, appurl)
                except Exception, e:
                    log.msg(_('Error calling function %s on parent server : %s') % ('verify_app_ticket',str(e)))
                    pass
        if not verified:
            log.msg("! %s %s !" % (_("Session verification failed for service"), get_service_from_url(appurl)))
        return verified, app_ticket

    def get_user_details(self, app_ticket, appurl, sections=False, renew=False, keep_valid=False):
        """vérifie un ticket et renvoie les informations sur l'utilisateur
        section : si True, on renvoie aussi une description de l'organisation des données
        """
        log.msg('--- %s %s' % (_("Validating session for"), appurl))
        code = 'INVALID_TICKET'
        detail = "%s : %s" % (_("Unknown ticket"), app_ticket)
        if app_ticket in self.app_sessions:
            log.msg('--- %s' % _("Session OK"))
            ticket = self.app_sessions[app_ticket]
            # si le paramètre renew a été envoyé et que le ticket provient d'une
            # session SSO, on ne le valide pas
            if not ticket.from_credentials and renew:
                detail = """Ticket %s was not delivered during authentication process""" % app_ticket
            elif ticket.verif_ticket(app_ticket, appurl, keep_valid):
                log.msg('--- %s' % _("Appurl verified"))
                return True, self._get_user_details(ticket.session_id, ticket.filter, sections)
            else:
                code = 'INVALID_SERVICE'
                detail = _("Ticket has not been delivered for service %s") % appurl
                log.msg("! %s !" % _("Error reading user data : invalid session"))
        else:
            # session inconnue sur ce serveur, on essaie sur le serveur parent
            if self._parent is not None:
                # récupération des infos sur le serveur parent
                # XXX  FIXME : re-filtrer les infos du serveur parent avec les filtres locaux ??
                try:
                    return self._parent.get_user_details(app_ticket, appurl, sections, renew)
                except Exception, e:
                    log.msg(_('Error calling function %s on parent server : %s') % ('get_user_details',str(e)))
                    pass
        log.msg("! %s %s !" % (_("User data access denied for service"), appurl))
        if sections:
            return False, ({'code':code,'detail':detail},"")
        return False, {'code':code,'detail':detail}

    def get_proxy_granting_ticket(self, app_ticket, service, pgturl):
        # si une url de proxy est donnée, on crée un ticket TGC
        # la conformité de l'url est vérifiée lors de la création.
        if app_ticket in self.app_sessions:
            ticket = self.app_sessions[app_ticket]
            if ticket.verif_ticket(app_ticket, service, keep_valid=True):
                ticket.generate_pgt(pgturl)
                self.proxy_granting_sessions[ticket.pgt] = ticket
                return ticket
        return None

    def get_proxy_ticket(self, pgt, target_service):
        # récupération de la session
        app_ticket = self.proxy_granting_sessions[pgt]
        appSessionID = gen_ticket_id('PT', self.address)
        # cas spécial : imapproxy ne fonctionne pas avec des pwd de + de 61 caractères
        if target_service.startswith('imap://') or target_service.startswith('imaps://'):
            appSessionID = appSessionID[:60]
        ticket = self.init_app_session(appSessionID, app_ticket.session_id, target_service, app_ticket.from_credentials)
        ticket.parent = app_ticket
        return ticket.ticket

    def invalidate_proxy(self, app_ticket):
        """Supprime les paramètres de proxy d'un app_ticket si les vérifications sur l'url de callback ont échoué
        """
        if app_ticket in self.app_sessions:
            ticket = self.app_sessions[app_ticket]
            if ticket.pgt is not None:
                del self.proxy_granting_sessions[ticket.pgt]
            ticket.reset_pgt()

    def _check_filter(self, appurl):
        """vérifie si un filtre est disponible pour l'url"""
        url = urlparse.urlparse(appurl)
        if url.scheme in self.apps:
            store = self.apps[url.scheme]
        else:
            store = self.apps['http']
        # recherche d'une application avec le même port/chemin
        if url.port in store:
            apps = store[url.port]
        else:
            # si pas de port correspondant, on regarde dans les applis sans port spécifique
            apps = store['all']
        # gestion du path, il faut que le path commence par ce qu'on veut
        # on trie les chemins connus pour commencer par le plus spécifique
        app_paths = apps.keys()
        app_paths.sort(reverse=True)
        url_path = url.path
        if not url_path.startswith('/'): url_path = '/' + url_path
        for baseurl in app_paths:
            check_baseurl = baseurl
            if not baseurl.startswith('/'): check_baseurl = '/' + baseurl
            if url_path.startswith(check_baseurl):
                scheme, addr, typeaddr, id_filter = apps[baseurl]
                if typeaddr == 'ip':
                    addr_ok = check_hostname_by_ip(url, addr)
                else:
                    addr_ok = check_hostname_by_domain(url, addr)
                if addr_ok:
                    # on a une appli correspondant à l'url, on renvoie le filtre associé
                    log.msg(_("Applied filter %s for url %s") % (id_filter, get_service_from_url(appurl)))
                    return id_filter
                break
        return "default"

    def _filter_data(self, infos, id_filter, sections):
        """filtrage en fonction de l'application si des filtres sont définis"""
        data = {}
        if id_filter in self.filters:
            for section, attrs in self.filters[id_filter].items():
                for libelle_cas, nom_val in attrs.items():
                    data[nom_val] = infos.get(nom_val, '')
            filter_data = self.filters[id_filter]
        else:
            # FIXME : définir les données à renvoyer par défaut !
            data.update(infos)
            filter_data = ""
        if sections:
            return data, filter_data
        else:
            return data

    def _get_user_details(self, session_id, id_filter, sections):
        """Renvoie les informations connues de l'utilisateur lié à <session_id>"""
        # on récupère l'id de l'utilisateur
        # user_id = self.user_sessions[session_id]
        if self.check_ticket_issuer(session_id):
            infos = self.data_cache[session_id]
        else:
            # session distante
            try:
                infos = self._parent.get_user_info(session_id,True)
            except Exception, e:
                log.msg(_('Error calling function %s on parent server : %s') % ('get_user_info',str(e)))
                infos = {}
        # ajout des informations statiques
        infos.update({'rne':[config.RNE],'nom_etab':[config.ETABLISSEMENT]})
        # lancement des fonctions de calcul de valeurs supplémentaires
        for varname, calc_func in self.user_infos.dict_infos.items():
            try:
                infos[varname] = calc_func(infos)
            except Exception, e:
                log.msg("""%s '%s' :  %s""" % (_("Error computing data for attribute"), varname, str(e)))
        infos = self._filter_data(infos, id_filter, sections)
        log.msg('--- %s' % _("User data sent"))
        return infos

